# 
# merge-points.sh: report where two branches have been merged
################################################################
# Copyright (C) 2001, 2002 Tom Lord
# 
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
# 

set -e 

################################################################
# special options
# 
# Some options are special:
# 
#       --version | -V
#       --help | -h
# 
if test $# -ne 0 ; then

  for opt in "$@" ; do
    case $opt in

      --version|-V) exec larch --version
                    ;;


      --help|-h)
                printf "report where two branches have been merged\\n"
                printf "usage: merge-points [options] INTO [FROM]\\n"
                printf "\\n"
                printf " -V --version                  print version info\\n"
                printf " -h --help                     display help\\n"
                printf "\\n"
                printf " -R --root root                specify the local archive root\\n"
                printf " -A --archive archive          specify the archive name\\n"
                printf "\\n"
                printf " -r --reverse                  list from most to least recent\\n"
		printf " -d --dir DIR                  instead of searching the archive,\\n"
		printf "                                 use the patch log in the tree\\n"
		printf "                                 containing DIR\\n"
                printf "\\n"
		printf " --full                        list included patches by full name\\n"
                printf "\\n"
                printf "Print a list of pairs of patch level names:\\n"
                printf "\\n"
                printf "	%%s\\\\t%%s   INTO-RVN  FROM-RVN\\n"
                printf "\\n"
                printf "where each pair indicates that at patch level INTO-RVN of\\n"
                printf "INTO, the patch log entry FROM-RVN was added.\\n"
                printf "\\n"
                printf "FROM may be a branch name, version name, or revision name.\\n"
		printf "If a branch or version name, all merges from that branch or version\\n"
		printf "are reported.  If a revision name, only the merge points for that\\n"
		printf "specific revision are reported.\\n"
                printf "\\n"
                printf "INTO may be a version name or revision name.  If a version name,\\n"
		printf "all merge points within that revision are printed.  If a revision\\n"
		printf "name, all merge points at that revision or earlier are printed.\\n"
                printf "\\n"
                printf "Output is sortd using patch-level ordering, primarilly by the\\n"
                printf "first column, and secondarily by the second column.\\n"
                printf "\\n"
                exit 0
                ;;

      *)
                ;;
    esac
  done
fi

################################################################
# Ordinary Options
# 
# 

archroot=
archive=
reverse=
reverse_r=
dir=
full=
debug_opt=

while test $# -ne 0 ; do

  case "$1" in 

    --debug)		shift
    			debug_opt=--debug
			printf "\n" 1>&2
			printf "merge-points: DEBUGGING ACTIVATED\n" 1>&2
			printf "\n" 1>&2
			set -x
			;;
			
    --full)	       shift
    		       full=--full
		       ;;

    -r|--reverse)      shift
    		       reverse=-r
		       reverse_r=r
		       ;;

    -d|--dir)          shift
                        if test $# -eq 0 ; then
                          printf "merge-points: -d and --dir require an argument\\n" 1>&2
                          printf "try --help\\n" 1>&2
                          exit 1
                        fi
                        dir="$1"
                        shift
                        ;;

    -R|--root)          shift
                        if test $# -eq 0 ; then
                          printf "merge-points: -R and --root require an argument\\n" 1>&2
                          printf "try --help\\n" 1>&2
                          exit 1
                        fi
                        archroot="$1"
                        shift
                        ;;

    -A|--archive)       shift
                        if test $# -eq 0 ; then
                          printf "merge-points: -A and --archive require an argument\\n" 1>&2
                          printf "try --help\\n" 1>&2
                          exit 1
                        fi
                        archive="$1"
                        shift
                        ;;

    --)			shift
    			break
			;;
			
    -*)                 printf "merge-points: unrecognized option (%s)\\n" "$1" 1>&2
                        printf "try --help\\n" 1>&2
                        exit 1
                        ;;

    *)                  break
                        ;;
  esac

done



################################################################
# Ordinary Arguments
# 

if test '(' $# -lt 1 ')' -o '(' $# -gt 2 ')' ; then
  printf "usage: merge-points [options] INTO [FROM]\\n" 1>&2
  printf "try --help\\n" 1>&2
  exit 1
fi

into="$1"
shift

if test $# -ne 0 ; then
  from="$1"
  shift
else 
  from=
fi

################################################################
# Sanity Check and Process Defaults
# 

here="`pwd`"

#
# validate DIR argument, if any
#


if test -z "$dir" ; then
  search_loc=archive
else
  search_loc=tree
  cd "$here"
  cd "$dir"
  wdroot="`larch tree-root --accurate`"
  dir="`pwd`/`basename \"$dir\"`"
fi


# 
# parse INTO
# 

larch valid-package-name -e merge-points --vsn --tolerant "$into"
into_archive="`larch parse-package-name -R \"$archroot\" -A \"$archive\" --arch \"$into\"`"
into_category="`larch parse-package-name --basename \"$into\"`"
into_branch="`larch parse-package-name \"$into\"`"
into_version="`larch parse-package-name --package-version \"$into\"`"

if larch valid-package-name --lvl "$into" ; then
  into_lvl="`larch parse-package-name --lvl \"$into\"`"
else
  into_lvl=
fi


#
# parse FROM
#

if test -z "$from" ; then

  from_prefix=
  full=--full

else

  if larch valid-archive-name "$from" ; then

    from_prefix="$from/"

  else

    larch valid-package-name -e merge-points --tolerant "$from" 
    from_archive="`larch parse-package-name -R \"$archroot\" -A \"$archive\" --arch \"$from\"`"

    from="$from_archive/`larch parse-package-name --non-arch \"$from\"`"

    from_category="`larch parse-package-name --basename \"$from\"`"
    from_branch="`larch parse-package-name \"$from\"`"

    if larch valid-package-name --tolerant --vsn "$from" ; then
      from_version="`larch parse-package-name --package-version \"$from\"`"
      from_prefix="$from_archive/$from_version--"
    else
      from_version=
      from_prefix="$from_archive/$from_branch--"
    fi

    if larch valid-package-name --lvl "$from" ; then
      from_lvl="`larch parse-package-name --lvl \"$from\"`"
      from_prefix="$from_prefix$from_lvl"
    else
      from_lvl=
    fi

  fi

fi


################################################################
# How to Translate One Log Entry to a Merge-Points Entry
# 
# 

entry()
{
  awk -v from_prefix="$from_prefix" \
      -v patch="$patch" \
      -v full="$full" \
    '
    function patch_logs_from_header(header)
    {
      n_patches = split(headers[header], patches, "[ \t]+");

      for (x = 1; x <= n_patches; ++x)
        {
    	  patch_prefix = substr (patches[x], 1, length (from_prefix));
    	  if (patch_prefix == from_prefix)
            {
	      from_lvl = patches[x];
	      if (full == "")
	        sub(".*--", "", from_lvl);
	      printf("%s\t%s\n", patch, from_lvl);
	    }
	}
    }

    function safe_getline()
    {
       status = getline;
       if (status < 0)
         {
	   print "I/O ERROR in valid-log-file" | "cat 1>&2";
	   exit (1);
	 }
       if (!status)
         eof_seen = 1;
       return status;
    }

    BEGIN { 
	     eof_seen = 0;

	     if (safe_getline())
	       {
                 while (!eof_seen && !match ($0, "^$"))
                   {
                     field = tolower ($1);
                     sub (":.*", "", field);
                     headers[field] = $0;
                     sub ("^[^:]*:[ \t]*", "", headers[field]);
                     while (safe_getline() && match ($0, "^[ \t]"))
                       {
                         headers[field] = headers[field] $0;
                       }
		   }
               }
           }

     { exit }
                  
     END {
	   patch_logs_from_header("new-patches");
         }'
}


################################################################
# Print the Merge-Points
# 

case $search_loc in

  archive)			patch_list="`larch revisions $debug_opt $reverse \"$into_archive/$into_version\"`"
  				log_catter=cat-archive-log
  				;;


  tree)				cd "$wdroot"
				patch_list="`larch log-ls $debug_opt $reverse \"$into_archive/$into_version\"`"
				log_catter=cat-log
				;;
esac


if test -z "$into_lvl" ; then
  skip=
else
  skip="$reverse"
fi

one_msg()
{
  if test -z "$full" ; then
    entry \
    | sed -e 's/-/ /g' \
    | sort -k "3,3$reverse_r" -k "4,4$reverse_r"n \
    | sed -e 's/\(.*\) \(.*\)	\(.*\) \(.*\)/\1-\2	\3-\4/'
  else
    entry \
    | sed -e 's/\(.*--\)\(.*\)-\(.*\)/\2 \3 \1/g' \
    | sort -k 3,3 -k "1,1$reverse_r" -k "2,2$reverse_r"n \
    | sed -e 's/\([^ ]*\) \([^ ]*\) \(.*\)/\3\1-\2/'
  fi
}

if test ! -z "$patch_list" ; then
  for patch in $patch_list ; do
  
    # when scanning backwards, skip patches until we find
    # $into_lvl
    #
    if test '(' "x$skip" != x ')' -a '(' "$patch" != "$into_lvl" ')' ; then
      continue
    fi
  
    skip=
  
    larch $log_catter $debug_opt "$into_archive/$into_version--$patch" | one_msg | pipe-friendly

  
    # when going forwards, stop after $into_lvl
    #
    if test '(' "x$reverse" = x ')' -a '(' "$patch" = "$into_lvl" ')' ; then
      break ;
    fi
  
  done
fi


# tag: Tom Lord Sun Jan  6 13:50:14 2002 (patch-logs/merge-points.sh)
#
